/* dataCell.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "dataCell.h"
#include "format.h"

#include <stdio.h>
#include <strstream.h>
#include <malloc.h>

string
enquote(string source) {
  // We have to double up any internal single quotes.
  int q = source.find('\'');
  while (q != string::npos) {
    source.insert(q, 1, '\'');
    q = source.find('\'', q+2);
  }

  return "'" + source + "'";
}

DataCell::
DataCell() {
  _type = DC_null;
}

DataCell::
DataCell(const DataCell &copy) : 
  _type(copy._type),
  _string_value(copy._string_value),
  _foreign_value(copy._foreign_value),
  _v(copy._v) 
{
}


// The type-conversion constructor.  This makes a new DataCell with
// the same value as the given DataCell, converted to the indicated
// type in some sensible way, if possible.  If the conversion cannot
// be made, the DataCell is left in its original type.
DataCell::
DataCell(const DataCell &copy, Type new_type) :
  _type(new_type),
  _string_value(copy._string_value),
  _foreign_value(copy._foreign_value),
  _v(copy._v) 
{
  if (_type == copy._type) {
    return;
  }

  switch (_type) {
  case DC_null:
    return;

  case DC_bool:
    switch (copy._type) {
    case DC_int:
      _v._bool_value = (copy._v._int_value != 0);
      return;

    case DC_float:
      _v._bool_value = (copy._v._float_value != 0);
      return;

    case DC_string:
      _v._bool_value = (!copy._string_value.empty() && 
			copy._string_value[0] != 'f');
      return;
    }
    break;

  case DC_int:
    switch (copy._type) {
    case DC_bool:
      _v._int_value = copy._v._bool_value ? 1 : 0;
      return;

    case DC_float:
    case DC_elapsed_time:
      _v._int_value = (int)copy._v._float_value;
      return;

    case DC_string:
      _v._int_value = atoi(copy._string_value.c_str());
      return;
    }
    break;

  case DC_float:
    switch (copy._type) {
    case DC_bool:
      _v._float_value = copy._v._bool_value ? 1.0 : 0.0;
      return;

    case DC_int:
      _v._float_value = (double)copy._v._int_value;
      return;

    case DC_elapsed_time:
      _v._float_value = copy._v._float_value;
      return;

    case DC_string:
      _v._float_value = atof(copy._string_value.c_str());
      return;
    }
    break;
    
  case DC_string:
    _string_value = copy.get_text();
    return;
    
  case DC_timestamp:
    break;

  case DC_elapsed_time:
    switch (copy._type) {
    case DC_int:
      _v._float_value = (double)copy._v._int_value;
      return;

    case DC_float:
      _v._float_value = copy._v._float_value;
      return;
    }
    break;
  }

  // Couldn't convert.
  _type = copy._type;
}
    

DataCell &DataCell::
operator = (const DataCell &copy) {
  _type = copy._type;
  _string_value = copy._string_value;
  _foreign_value = copy._foreign_value;
  _v = copy._v;
  return *this;
}

DataCell::
DataCell(int value, bool valid) {
  if (valid) {
    _v._int_value = value;
    _type = DC_int;
  } else {
    _type = DC_null;
  }
}

DataCell::
DataCell(double value, bool valid) {
  if (valid) {
    _v._float_value = value;
    _type = DC_float;
  } else {
    _type = DC_null;
  }
}

DataCell::
DataCell(const string &value, bool valid) {
  if (valid) {
    _string_value = value;
    _type = DC_string;
  } else {
    _type = DC_null;
  }
}

DataCellBool::
DataCellBool(bool value, bool valid) {
  if (valid) {
    _v._bool_value = value;
    _type = DC_bool;
  } else {
    _type = DC_null;
  }
}

DataCellTimestamp::
DataCellTimestamp(time_t value, bool valid) {
  if (valid) {
    _v._time_value = value;
    _type = DC_timestamp;
  } else {
    _type = DC_null;
  }
}

DataCellElapsedTime::
DataCellElapsedTime(double value, bool valid) {
  if (valid) {
    _v._float_value = value;
    _type = DC_elapsed_time;
  } else {
    _type = DC_null;
  }
}

DataCell::
DataCell(const TableDataProxy &proxy, bool valid) {
  if (valid) {
    _foreign_value = proxy;
    _type = DC_foreign_key;
  } else {
    _type = DC_null;
  }
}

DataCell::
DataCell(const char *db_value, Type type) {
  // If the db_value is the NULL string, or it's the empty string
  // (except for type DC_string), we consider the value to be NULL.
  if (db_value == NULL ||
      (db_value[0] == '\0' && type != DC_string)) {
    _type = DC_null;
    return;
  }

  _type = type;

  switch (type) {
  case DC_bool:
    _v._bool_value = (db_value[0] == 't');
    return;
    
  case DC_int:
    _v._int_value = atoi(db_value);
    return;
    
  case DC_float:
  case DC_elapsed_time:
    _v._float_value = atof(db_value);
    return;
    
  case DC_string:
    _string_value = db_value;
    return;
    
  case DC_timestamp:
    {
      int year, mon, mday;
      int hour, min, sec;

      int result =
	sscanf(db_value, "%04d-%02d-%02d %02d:%02d:%02d",
	       &year, &mon, &mday, &hour, &min, &sec);
      if (result != 6) {
	result =
	  sscanf(db_value, "%04d%02d%02d%02d%02d%02d",
		 &year, &mon, &mday, &hour, &min, &sec);
      }
      if (result != 6) {
	cerr << "Unable to decode time " << db_value << "\n";
	exit(1);
      }

      struct tm time;
      time.tm_year = year - 1900;
      time.tm_mon = mon - 1;
      time.tm_mday = mday;
      time.tm_hour = hour;
      time.tm_min = min;
      time.tm_sec = sec;
      time.tm_isdst = false;
      
      // mktime() normally accepts a struct tm which is defined in the
      // local timezone and returns a time_t in universal time.  However,
      // we are giving it a struct tm which is itself in universal time.
      // Thus, the conversion will be wrong by the amount of the timezone
      // difference.
      
      // Fortunately, mktime() also sets the global variable timezone to
      // be this timezone difference.  Thus, we only have to subtract it
      // back out again to get the correct answer.
      time_t t = mktime(&time);
      _v._time_value = t - timezone;
      return;
    }
  }

  cerr << "Unexpected type in DataCell formatted constructor: " 
       << _type << " (value is " << (*this) << ")\n";
  abort();
}

bool DataCell::
operator == (const DataCell &other) const {
  if (_type != other._type) {
    // Generally, if the type is different, the values are different.

    // As a special exception, however, integers are allowed to stand
    // in for foreign keys.
    if (_type == DC_int && other._type == DC_foreign_key) {
      return (_v._int_value == other._foreign_value.get_primary_key());
    } else if (_type == DC_foreign_key && other._type == DC_int) {
      return (other._v._int_value == _foreign_value.get_primary_key());
    }

    // Similarly, NULL is allowed to match an empty foreign key.
    if (_type == DC_null && other._type == DC_foreign_key) {
      return (!other._foreign_value.has_table());
    } else if (_type == DC_foreign_key && other._type == DC_null) {
      return (!_foreign_value.has_table());
    }
    
    return false;
  }

  switch (_type) {
    case DC_null:
    // All null values are equivalent.
    return true;

    case DC_int:
    return _v._int_value == other._v._int_value;

    case DC_float:
    case DC_elapsed_time:
    return _v._float_value == other._v._float_value;

    case DC_bool:
    return _v._bool_value == other._v._bool_value;

    case DC_timestamp:
    return _v._time_value == other._v._time_value;

    case DC_string:
    return _string_value == other._string_value;

    case DC_foreign_key:
    // We consider two NULL foreign key references to be the same.
    return (_foreign_value == other._foreign_value ||
	    !_foreign_value.has_table() && !other._foreign_value.has_table());
    
  }

  cerr << "Unexpected type in DataCell::operator ==(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return false;
}

bool DataCell::
operator != (const DataCell &other) const {
  return !(operator == (other));
}

void DataCell::
clear() {
  _type = DC_null;
}

// Reintreprets the DataCell as the foreign key to a record in another
// table.  Returns the new DataCell that corresponds to the foreign
// record.
DataCell DataCell::
as_foreign_key(TableID table_id) const {
  switch (_type) {
  case DC_null:
    return DataCell();

  case DC_int:
    return DataCell(TableDataProxy(table_id, _v._int_value));
  }

  cerr << "Unexpected type in DataCell::as_foreign_key(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return DataCell();
}

// record_db() forces a write of the foreign record to the database,
// if the DataCell type is DC_foreign_key, and a valid foreign record
// actually exists.  Otherwise it does nothing.
void DataCell::
record_db() {
  if (_type == DC_foreign_key) {
    if (_foreign_value.has_table()) {
      _foreign_value.record_db();
    }
  }
}


string DataCell::
q_str() const {
  switch (_type) {
  case DC_null:
    return "NULL";

  case DC_bool:
    if (_v._bool_value) {
      return string("'t'");
    } else {
      return string("'f'");
    }

  case DC_int:
    return format_int(_v._int_value);

  case DC_float:
  case DC_elapsed_time:
    return format_float(_v._float_value);

  case DC_string:
    return enquote(_string_value);

  case DC_timestamp:
    {
      char time_str[128];
      struct tm *time = gmtime(&_v._time_value);
      sprintf(time_str,
	      "'%04d-%02d-%02d %02d:%02d:%02d+00'",
	      time->tm_year + 1900, time->tm_mon + 1, time->tm_mday,
	      time->tm_hour, time->tm_min, time->tm_sec);
      return time_str;
    }

  case DC_foreign_key:
    if (_foreign_value.has_table() && 
	_foreign_value.get_primary_key() >= 0) {
      return format_int(_foreign_value.get_primary_key());
    } else {
      return "NULL";
    }
  }

  cerr << "Unexpected type in DataCell::q_str(): " << _type << "\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return string();
}

string DataCell::
e_str() const {
  if (is_null()) {
    return " IS NULL";
  } else {
    return "=" + q_str();
  }
}

bool DataCell::
is_null() const {
  return _type == DC_null;
}

DataCell::Type DataCell::
get_type() const {
  return _type;
}

bool DataCell::
get_bool(bool null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_bool:
    return _v._bool_value;

  case DC_int:
    return _v._int_value != 0;

  case DC_float:
  case DC_elapsed_time:
    return _v._float_value != 0;
  }    

  cerr << "Unexpected type in DataCell::get_bool(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

int DataCell::
get_int(int null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_int:
    return _v._int_value;
  }    

  cerr << "Unexpected type in DataCell::get_int(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

double DataCell::
get_float(double null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_int:
    return _v._int_value;

  case DC_float:
  case DC_elapsed_time:
    return _v._float_value;
  }    

  cerr << "Unexpected type in DataCell::get_float(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

const string &DataCell::
get_string(const string &null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_string:
    return _string_value;
  }    

  cerr << "Unexpected type in DataCell::get_string(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

string DataCell::
get_text(const string &null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_bool:
    if (_v._bool_value) {
      return string("t");
    } else {
      return string("f");
    }

  case DC_int:
    return format_int(_v._int_value);

  case DC_float:
    return format_float(_v._float_value);

  case DC_string:
    return _string_value;

  case DC_timestamp:
    return format_datetime(_v._time_value);

  case DC_elapsed_time:
    return format_elapsed_time(_v._float_value);

  case DC_foreign_key:
    if (_foreign_value.has_table() && 
	_foreign_value.get_primary_key() >= 0) {
      return get_table_name(_foreign_value.get_table_id()) + "(" +
	format_int(_foreign_value.get_primary_key()) + ")";
    } else {
      return null_value;
    }
  }    

  cerr << "Unexpected type in DataCell::get_text(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

time_t DataCell::
get_time(time_t null_value) const {
  switch (_type) {
  case DC_null:
    return null_value;

  case DC_timestamp:
    return _v._time_value;
  }    

  cerr << "Unexpected type in DataCell::get_time(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return null_value;
}

TableDataProxy DataCell::
get_foreign(TableID table_id) const {
  switch (_type) {
  case DC_null:
    return TableDataProxy(table_id);

  case DC_int:
    return TableDataProxy(table_id, _v._int_value);

  case DC_foreign_key:
    return _foreign_value;
  }    

  cerr << "Unexpected type in DataCell::get_foreign(): " << _type 
       << " (value is " << (*this) << ")\n";
  abort();

  // This is just to keep the compiler happy.  We'll never get here.
  return TableDataProxy(table_id);
}

void DataCell::
output(ostream &out) const {
  out << get_text() << " (" << get_type() << ")";
}


ostream &
operator << (ostream &out, DataCell::Type type) {
  switch (type) {
  case DataCell::DC_null:
    return out << "null";

  case DataCell::DC_bool:
    return out << "bool";

  case DataCell::DC_int:
    return out << "int";

  case DataCell::DC_float:
    return out << "float";
    
  case DataCell::DC_string:
    return out << "string";
    
  case DataCell::DC_timestamp:
    return out << "timestamp";

  case DataCell::DC_elapsed_time:
    return out << "elapsed_time";

  case DataCell::DC_foreign_key:
    return out << "foreign";
    
  default:
    return out << "**invalid type (" << (int)type << ")**"; 
  }
}
    
